@@ -2,7 +2,7 @@
    Originally created by Burt Peterson
    Updated and maintained by Dustin Roeder (dmroeder@gmail.com)
 
-   Copyright 2019 Dustin Roeder
+   Copyright 2020 Dustin Roeder
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
@@ -26,7 +26,6 @@ from .lgxDevice import LGXDevice, GetDevice, GetVendor
 from random import randrange
 from struct import pack, unpack_from
 
-
 class PLC:
 
     def __init__(self, ip_address="", slot=0):
@@ -36,6 +35,7 @@ class PLC:
         self.IPAddress = ip_address
         self.ProcessorSlot = slot
         self.Micro800 = False
+        self.Route = None
         self.Port = 44818
         self.VendorID = 0x1337
         self.Context = 0x00
@@ -137,10 +137,7 @@ class PLC:
 
         returns Response class (.TagName, .Value, .Status)
         """
-        tag_list = self._getTagList(allTags)
-        updated_list = self._getUDT(tag_list.Value)
-
-        return Response(None, updated_list, tag_list.Status)
+        return self._getTagList(allTags)
 
     def GetProgramTagList(self, programName):
         """
@@ -149,6 +146,9 @@ class PLC:
 
         returns Response class (.TagName, .Value, .Status)
         """
+        conn = self._connect()
+        if not conn[0]:
+            return Response(programName, None, conn[1])
 
         # If ProgramNames is empty then _getTagList hasn't been called
         if not self.ProgramNames:
@@ -163,7 +163,7 @@ class PLC:
             program_tags = self._getUDT(program_tags.Value)
             return Response(None, program_tags, status)
         else:
-            return Response(None, None, 'Program not found, please check name!')
+            return Response(programName, None, 'Program not found, please check name!')
 
     def GetProgramsList(self):
         """
@@ -173,6 +173,11 @@ class PLC:
 
         returns Response class (.TagName, .Value, .Status)
         """
+
+        conn = self._connect()
+        if not conn[0]:
+            return Response(None, None, conn[1])
+
         tags = ''
         if not self.ProgramNames:
             tags = self._getTagList(False)
@@ -217,12 +222,13 @@ class PLC:
 
     def _readTag(self, tag_name, elements, data_type):
         """
-        processes the read request
+        Processes the read request
         """
         self.Offset = 0
 
-        if not self._connect():
-            return None
+        conn = self._connect()
+        if not conn[0]:
+            return Response(tag_name, None, conn[1])
 
         tag, base_tag, index = _parseTagName(tag_name, 0)
         resp = self._initial_read(tag, base_tag, data_type)
@@ -266,8 +272,9 @@ class PLC:
         self.Offset = 0
         write_data = []
         
-        if not self._connect():
-            return None
+        conn = self._connect()
+        if not conn[0]:
+            return Response(tag_name, value, conn[1])
 
         tag, base_tag, index = _parseTagName(tag_name, 0)
         resp = self._initial_read(tag, base_tag, data_type)
@@ -329,8 +336,9 @@ class PLC:
         tag_count = len(tags)
         self.Offset = 0
 
-        if not self._connect():
-            return None
+        conn = self._connect()
+        if not conn[0]:
+            return [Response(None, None, conn[1])]
 
         for tag in tags:
             if isinstance(tag, (list, tuple)):
@@ -380,8 +388,9 @@ class PLC:
         tag_count = len(write_data)
         self.Offset = 0
 
-        if not self._connect():
-            return None
+        conn = self._connect()
+        if not conn[0]:
+            return [Response(None, value, conn[1])]
 
         for wd in write_data:
 
@@ -426,16 +435,15 @@ class PLC:
         eip_header = self._buildEIPHeader(request)
         status, ret_data = self._getBytes(eip_header)
 
-        #tags = [t[0] for t in stuff]
-
         return self._multiWriteParser(write_data, ret_data)
 
     def _getPLCTime(self, raw=False):
         """
         Requests the PLC clock time
         """
-        if not self._connect():
-            return None
+        conn = self._connect()
+        if not conn[0]:
+            return Response(None, None, conn[1])
 
         AttributeService = 0x03
         AttributeSize = 0x02
@@ -475,8 +483,9 @@ class PLC:
         """
         Requests the PLC clock time
         """
-        if not self._connect():
-            return None
+        conn = self._connect()
+        if not conn[0]:
+            return Response(None, None, conn[1])
 
         AttributeService = 0x04
         AttributeSize = 0x02
@@ -507,8 +516,9 @@ class PLC:
         """
         Requests the controller tag list and returns a list of LgxTag type
         """
-        if not self._connect():
-            return None
+        conn = self._connect()
+        if not conn[0]:
+            return Response(None, None, conn[1])
 
         self.Offset = 0
         tags = []
@@ -553,15 +563,17 @@ class PLC:
                         tags += self._extractTagPacket(ret_data, program_name)
                     else:
                         return Response(None, None, status)
-
-        return Response(None, tags, status)
+                        
+        updated_list = self._getUDT(tags)
+        return Response(None, updated_list, status)
 
     def _getProgramTagList(self, programName):
         """
         Requests tag list for a specific program and returns a list of LgxTag type
         """
-        if not self._connect():
-            return None
+        conn = self._connect()
+        if not conn[0]:
+            return Response(None, None, conn[1])
 
         self.Offset = 0
         tags = []
@@ -584,6 +596,7 @@ class PLC:
             else:
                 return Response(None, None, status)
 
+        updated_list = self._getUDT(tags)
         return Response(None, tags, status)
 
     def _getUDT(self, tag_list):
@@ -629,10 +642,6 @@ class PLC:
         """
         Get the attributes of a UDT
         """
-
-        if not self._connect():
-            return None
-
         request = self._buildTemplateAttributes(instance)
         eip_header = self._buildEIPHeader(request)
         status, ret_data = self._getBytes(eip_header)
@@ -642,16 +651,16 @@ class PLC:
         """
         Get the members of a UDT so we can get it
         """
-        if not self._connect():
-            return None
-
         request = self._readTemplateService(instance, dataLen)
         eip_header = self._buildEIPHeader(request)
         status, ret_data = self._getBytes(eip_header)
         return ret_data
 
     def _buildTemplateAttributes(self, instance):
-
+        """
+        Build the template attribute packet, part of
+        retreiving the UDT names
+        """
         TemplateService = 0x03
         TemplateLength = 0x03
         TemplateClassType = 0x20
@@ -678,7 +687,10 @@ class PLC:
                     Attrib1)
 
     def _readTemplateService(self, instance, dataLen):
-
+        """
+        Build the template attribute packet, part of
+        retreiving the UDT names
+        """
         TemplateService = 0x4c
         TemplateLength = 0x03
         TemplateClassType = 0x20
@@ -699,6 +711,10 @@ class PLC:
                     DataLength)
 
     def _discover(self):
+        """
+        Discover devices on the network, similar to the RSLinx
+        Ethernet I/P driver
+        """
         devices = []
         request = self._buildListIdentity()
 
@@ -752,51 +768,9 @@ class PLC:
         Request the properties of a module in a particular
         slot.  Returns LgxDevice
         """
-        if not self._connect(False):
-            return None
-
-        AttributeService = 0x01
-        AttributeSize = 0x02
-        AttributeClassType = 0x20
-        AttributeClass = 0x01
-        AttributeInstanceType = 0x24
-        AttributeInstance = 0x01
-        PathRouteSize = 0x01
-        Reserved = 0x00
-        Backplane = 0x01
-        LinkAddress = slot
-
-        AttributePacket = pack('<10B',
-                               AttributeService,
-                               AttributeSize,
-                               AttributeClassType,
-                               AttributeClass,
-                               AttributeInstanceType,
-                               AttributeInstance,
-                               PathRouteSize,
-                               Reserved,
-                               Backplane,
-                               LinkAddress)
-
-        frame = self._buildCIPUnconnectedSend() + AttributePacket
-        eip_header = self._buildEIPSendRRDataHeader(len(frame)) + frame
-        pad = pack('<I', 0x00)
-        self.Socket.send(eip_header)
-        ret_data = pad + self.recv_data()
-        status = unpack_from('<B', ret_data, 46)[0]
-
-        if status == 0:
-            return Response(None, _parseIdentityResponse(ret_data), status)
-        else:
-            return Response(None, LGXDevice(), status)
-
-    def _getDeviceProperties(self):
-        """
-        Request the properties of a device at the
-        specified IP address.  Returns LgxDevice
-        """
-        if not self._connect(False):
-            return None
+        conn = self._connect(False)
+        if not conn[0]:
+            return Response(None, LGXDevice(), conn[1])
 
         AttributeService = 0x01
         AttributeSize = 0x02
@@ -813,7 +787,47 @@ class PLC:
                                AttributeInstanceType,
                                AttributeInstance)
 
-        frame = AttributePacket
+        ConnectionPath = self._unconnectedPath(slot)
+
+        frame = self._buildCIPUnconnectedSend() + AttributePacket + ConnectionPath
+        eip_header = self._buildEIPSendRRDataHeader(len(frame)) + frame
+        pad = pack('<I', 0x00)
+        self.Socket.send(eip_header)
+        ret_data = pad + self.recv_data()
+        status = unpack_from('<B', ret_data, 46)[0]
+
+        if status == 0:
+            return Response(None, _parseIdentityResponse(ret_data), status)
+        else:
+            return Response(None, LGXDevice(), status)
+
+    def _getDeviceProperties(self):
+        """
+        Request the properties of a device at the
+        specified IP address.  Returns LgxDevice
+        """
+        conn = self._connect(False)
+        if not conn[0]:
+            return Response(None, LGXDevice(), conn[1])
+
+        AttributeService = 0x01
+        AttributeSize = 0x02
+        AttributeClassType = 0x20
+        AttributeClass = 0x01
+        AttributeInstanceType = 0x24
+        AttributeInstance = 0x01
+
+        AttributePacket = pack('<6B',
+                               AttributeService,
+                               AttributeSize,
+                               AttributeClassType,
+                               AttributeClass,
+                               AttributeInstanceType,
+                               AttributeInstance)
+
+        ConnectionPath = self._unconnectedPath(slot=0)
+
+        frame = self._buildCIPUnconnectedSend() + AttributePacket + ConnectionPath
         eip_header = self._buildEIPSendRRDataHeader(len(frame)) + frame
         pad = pack('<I', 0x00)
         self.Socket.send(eip_header)
@@ -837,22 +851,23 @@ class PLC:
                 # connection type changed, need to close so we can reconnect
                 self._closeConnection()
             else:
-                return True
+                return (True, 'Success')
 
         # Make sure the connection size is correct
         if not 500 <= self.ConnectionSize <= 4000:
-            raise ValueError("ConnectionSize must be an integer between 500 and 4000")
+            return (False, 'ConnectionSize must be an integer between 500 and 4000')
 
         try:
             self.Socket = socket.socket()
             self.Socket.settimeout(5.0)
             self.Socket.connect((self.IPAddress, self.Port))
-        except(socket.error):
+        except socket.error as e:
             self.SocketConnected = False
             self.SequenceCounter = 1
             self.Socket.close()
-            raise
+            return (False, e)
 
+        # register the session
         self.Socket.send(self._buildRegisterSession())
         ret_data = self.recv_data()
         if ret_data:
@@ -860,33 +875,36 @@ class PLC:
             self._registered = True
         else:
             self.SocketConnected = False
-            raise Exception("Failed to register session")
+            return (False, 'Register session failed')
 
+        # forward open connection
         if connected:
             self.Socket.send(self._buildForwardOpenPacket())
-            ret_data = self.recv_data()
+            try:
+                ret_data = self.recv_data()
+            except socket.timeout as e:
+                return (False, e)
             sts = unpack_from('<b', ret_data, 42)[0]
             if not sts:
                 self.OTNetworkConnectionID = unpack_from('<I', ret_data, 44)[0]
                 self._connected = True
             else:
                 self.SocketConnected = False
-                raise Exception("Forward Open Failed")
+                return (False, 'Forward open failed')
+
         self.SocketConnected = True
-        return self.SocketConnected
+        return (self.SocketConnected, 'Success')
 
     def _closeConnection(self):
         """
         Close the connection to the PLC (forward close, unregister session)
         """
         self.SocketConnected = False
-        
-        
         try:
-            if self._registered:
+            if self._connected:
                 close_packet = self._buildForwardClosePacket()
                 self.Socket.send(close_packet)
-            if self._connected:
+            if self._registered:
                 unreg_packet = self._buildUnregisterSession()
                 self.Socket.send(unreg_packet)
             self.Socket.close()
@@ -939,6 +957,9 @@ class PLC:
                     EIPOptionFlag)
 
     def _buildUnregisterSession(self):
+        """
+        Build Unregister session
+        """
         EIPCommand = 0x66
         EIPLength = 0x0
         EIPSessionHandle = self.SessionHandle
@@ -1033,16 +1054,11 @@ class PLC:
                            CIPTransportTrigger)
 
         # add the connection path
-        if self.Micro800:
-            ConnectionPath = [0x20, 0x02, 0x24, 0x01]
-        else:
-            ConnectionPath = [0x01, self.ProcessorSlot, 0x20, 0x02, 0x24, 0x01]
 
-        ConnectionPathSize = int(len(ConnectionPath)/2)
-        pack_format = '<B' + str(len(ConnectionPath)) + 'B'
-        CIPConnectionPath = pack(pack_format, ConnectionPathSize, *ConnectionPath)
-
-        return ForwardOpen + CIPConnectionPath
+        path_size, path = self._connectedPath()
+        connection_path = pack('<B', path_size)
+        connection_path += path
+        return ForwardOpen + connection_path
 
     def _buildForwardClose(self):
         """
@@ -1075,18 +1091,15 @@ class PLC:
                             CIPOriginatorSerialNumber)
 
         # add the connection path
-        if self.Micro800:
-            ConnectionPath = [0x20, 0x02, 0x24, 0x01]
-        else:
-            ConnectionPath = [0x01, self.ProcessorSlot, 0x20, 0x02, 0x24, 0x01]
-
-        ConnectionPathSize = int(len(ConnectionPath)/2)
-        pack_format = '<H' + str(len(ConnectionPath)) + 'B'
-        CIPConnectionPath = pack(pack_format, ConnectionPathSize, *ConnectionPath)
-
-        return ForwardClose + CIPConnectionPath
+        path_size, path = self._connectedPath()
+        connection_path = pack('<BB', path_size, 0x00)
+        connection_path += path
+        return ForwardClose + connection_path
 
     def _buildEIPSendRRDataHeader(self, frameLen):
+        """
+        Build the EIP Send RR Data Header
+        """
         EIPCommand = 0x6F
         EIPLength = 16+frameLen
         EIPSessionHandle = self.SessionHandle
@@ -1144,6 +1157,76 @@ class PLC:
                     CIPTimeoutTicks,
                     ServiceSize)
 
+    def _connectedPath(self):
+        """
+        Build the connected path porition of the packet
+        """
+        # if a route was provided, use it, otherwise use
+        # the default route
+        if self.Route:
+            route = self.Route
+        else:
+            route = [(0x01, self.ProcessorSlot)]
+
+        path = []
+        if self.Micro800:
+            pass
+        else:
+            for segment in route:
+                if isinstance(segment[1], int):
+                    # port segment
+                    path += segment
+                else:
+                    # port segment with link
+                    path.append(segment[0]+0x10)
+                    path.append(len(segment[1]))
+                    for c in segment[1]:
+                        path.append(ord(c))
+                    # byte align
+                    if len(path)%2:
+                        path.append(0x00)
+
+        path += [0x20, 0x02, 0x24, 0x01]
+
+        path_size = int(len(path)/2)
+        pack_format = '<{}B'.format(len(path))
+        connection_path = pack(pack_format, *path)
+
+        return path_size, connection_path
+
+    def _unconnectedPath(self, slot):
+        """
+        Build the unconnection path portion of the packet
+        """
+        # if a route was provided, use it, otherwise use
+        # the default route
+        if self.Route:
+            route = self.Route
+        else:
+            route = [(0x01, slot)]
+
+        reserved = 0x00
+        path = []
+        for segment in route:
+            if isinstance(segment[1], int):
+                # port segment
+                path += segment
+            else:
+                # port segment with link
+                path.append(segment[0]+0x10)
+                path.append(len(segment[1]))
+                for c in segment[1]:
+                    path.append(ord(c))
+                # byte align
+                if len(path)%2:
+                    path.append(0x00)
+
+        path_size = int(len(path)/2)
+        pack_format = '<{}BBB'.format(len(path))
+        connection_path = pack(pack_format, path_size, reserved, *path)
+
+        return connection_path
+
     def _buildTagIOI(self, tagName, data_type):
         """
         The tag IOI is basically the tag name assembled into
@@ -1582,12 +1665,12 @@ class PLC:
             return tag, None, status
 
     def _convert_write_data(self, tag, data_type, write_values):
-        '''
+        """
         In order to handle write requests that are larger than a single
         packet, we'll break up the values to write into multiple lists
         of values.  The size of each list will be calculated based on the
         connection size, length of the tag name and the data type.
-        '''
+        """
         # packet header is always 110 bytes
         packet_overhead = 110
         # calculate number of bytes tag name will occupy
@@ -1692,9 +1775,6 @@ class PLC:
 
         return reply
 
-        
-
-
     def _buildListIdentity(self):
         """
         Build the list identity request for discovering Ethernet I/P
@@ -1770,7 +1850,6 @@ class PLC:
                 work.append(0x00)
         return work
 
-
 def _getBitOfWord(tag, value):
     """
     Takes a tag name, gets the bit from the end of
@@ -1793,7 +1872,6 @@ def _getBitOfWord(tag, value):
             pass
     return returnValue
 
-
 def _getWordCount(start, length, bits):
     """
     Get the number of words that the requested
@@ -1807,7 +1885,6 @@ def _getWordCount(start, length, bits):
     totalWords = (newEnd-1) / bits
     return totalWords + 1
 
-
 def _parseTagName(tag, offset):
     """
     parse the packet to get the base tag name
@@ -1837,7 +1914,6 @@ def _parseTagName(tag, offset):
     except Exception:
         return tag, bt, 0
 
-
 def BitofWord(tag):
     """
     Test if the user is trying to write to a bit of a word
@@ -1849,7 +1925,6 @@ def BitofWord(tag):
     else:
         return False
 
-
 def BitValue(value, bitno):
     """
     Returns the specific bit of a words value
@@ -1860,7 +1935,6 @@ def BitValue(value, bitno):
     else:
         return False
 
-
 def _parseIdentityResponse(data):
     # we're going to take the packet and parse all
     #  the data that is in it.
@@ -1893,7 +1967,6 @@ def _parseIdentityResponse(data):
 
     return resp
 
-
 def parseLgxTag(packet, programName):
 
     t = LgxTag()
@@ -1973,7 +2046,6 @@ class Response:
 
         return '{} {} {}'.format(self.TagName, self.Value, self.Status)
 
-
 def get_error_code(status):
     """
     Get the CIP error code string, if the status is a string it will be returned
@@ -1994,7 +2066,6 @@ def get_error_code(status):
         err = 'Unknown error {}'.format(status)
     return err
 
-
 cip_error_codes = {0x00: 'Success',
                    0x01: 'Connection failure',
                    0x02: 'Resource unavailable',
